/*
 * $Header$
 *
 * Copyright (c) 2002  Tatsuo Ishii
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above
 * copyright notice and this paragraph and the following two
 * paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS
 * IS" BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include "postgres.h"

#include "fmgr.h"
#include "access/genam.h"
#include "access/heapam.h"
#include "access/transam.h"
#include "utils/syscache.h"
#include "catalog/namespace.h"
#include "utils/builtins.h"

PG_FUNCTION_INFO_V1(pgstatindex);

extern Datum pgstatindex(PG_FUNCTION_ARGS);

/* ----------
 * pgstatindex:
 * returns the percentage of dead index tuples
 * XXX Fix me: this function only works with btree and hash indexes.
 * Probably would not work with rtree and gist indexes (i have not tested
 * yet though).
 *
 * C FUNCTION definition
 * pgstatindex(TEXT) returns FLOAT8
 * ----------
 */
Datum
pgstatindex(PG_FUNCTION_ARGS)
{
	static Oid IndexGetRelation(Oid indexId);	/* get parent relation oid */

	text	   *relname = PG_GETARG_TEXT_P(0);
	double		table_len;
	uint64		tuple_len = 0;
	uint64		dead_tuple_len = 0;
	uint64		tuple_count = 0;
	uint64		dead_tuple_count = 0;
	double		tuple_percent;
	double		dead_tuple_percent;
	uint64		free_space = 0; /* free/reusable space in bytes */
	double		free_percent;	/* free/reusable space in % */

	RangeVar   *relrv;
	Relation rel;		/* index relation */
	IndexScanDesc scan;	/* index scan descriptor */
	BlockNumber nblocks;		/* number of total blocks for this index */
	BlockNumber block = 0;		/* next block to count free space in */
	Relation heap = InvalidRelation ;	/* parent heap relation */
	Buffer buffer;

	/* open specified index */
	relrv = makeRangeVarFromNameList(textToQualifiedNameList(relname,
														"pgstatindex"));
	rel = index_openrv(relrv);

	/* get total number of blocks for this index */
	nblocks = RelationGetNumberOfBlocks(rel);

	/* start the scan */
	scan = index_beginscan(NULL, rel, NULL, 0 , NULL);
	/* we need to see dead index tuples */
	scan->ignore_killed_tuples = false;

	/* scan over each index tuple */
	while (index_getnext_indexitem(scan, ForwardScanDirection))
	{
		Page		page;
		BlockNumber blkno;		/* block number for this index page */
		OffsetNumber offnum;	/* page offset for this index page */
		IndexTuple	itup;		/* index tuple */
		HeapTupleData tuple;	/* parent relation tuple */
		bool is_deleted = false;		/* is the index tuple deleted? */

		/* get current block number */
		blkno = ItemPointerGetBlockNumber(&scan->currentItemData);
		/* get buffer cache for current block */
		buffer = ReadBuffer(rel, blkno);
		/* get tuple offset number in current page */
		offnum = ItemPointerGetOffsetNumber(&scan->currentItemData);
		page = BufferGetPage(buffer);
		/* get index tuple */
		itup = (IndexTuple)PageGetItem(page, PageGetItemId(page, offnum));
		/* get parent relation TID */
		tuple.t_self = itup->t_tid;

		/* has the index tuple been deleted? */
		if (ItemIdDeleted(PageGetItemId(page, offnum)))
		{
			is_deleted = true;
			ReleaseBuffer(buffer);
		}
		else
		{
			ReleaseBuffer(buffer);
			/*
			 * check to see if the parent tuple is valid
			 */
			/* open the parent relation if not opened yet */
			if (!RelationIsValid(heap))
			{
				heap = heap_open(IndexGetRelation(rel->rd_id), AccessShareLock);
			}
			/* get parent relation tuple */
			if (!heap_fetch(heap, SnapshotNow, &tuple, &buffer, false, NULL))
			{
				is_deleted = true;
			}
			ReleaseBuffer(buffer);
		}

		if (!is_deleted)
		{
			tuple_len += IndexTupleSize(itup);
			tuple_count++;
		}
		else
		{
			dead_tuple_len += IndexTupleSize(itup);
			dead_tuple_count++;
		}

		/*
		 * To avoid physically reading the table twice, try to do the
		 * free-space scan in parallel with the heap scan.  However,
		 * heap_getnext may find no tuples on a given page, so we cannot
		 * simply examine the pages returned by the heap scan.
		 */

		while (block <= blkno)
		{
			buffer = ReadBuffer(rel, block);
			free_space += PageGetFreeSpace((Page) BufferGetPage(buffer));
			ReleaseBuffer(buffer);
			block++;
		}
	}

	/* end the index scan */
	index_endscan(scan);

	while (block < nblocks) {
		buffer = ReadBuffer(rel, block);
		free_space += PageGetFreeSpace((Page) BufferGetPage(buffer));
		ReleaseBuffer(buffer);
		block++;
	}

	index_close(rel);	/* close the index */

	/* close the parent heap relation */
	if (RelationIsValid(heap))
		heap_close(heap, AccessShareLock);

	table_len = (double) nblocks *BLCKSZ;

	if (nblocks == 0) {
		tuple_percent = 0.0;
		dead_tuple_percent = 0.0;
		free_percent = 0.0;
	}
	else {
		tuple_percent = (double) tuple_len *100.0 / table_len;
		dead_tuple_percent = (double) dead_tuple_len *100.0 / table_len;
		free_percent = (double) free_space *100.0 / table_len;
	}

	elog(NOTICE, "physical length: %.2fMB live tuples: %.0f (%.2fMB, %.2f%%) dead tuples: %.0f (%.2fMB, %.2f%%) free/reusable space: %.2fMB (%.2f%%) overhead: %.2f%%",

		 table_len / (1024 * 1024),		/* physical length in MB */

		 (double) tuple_count,	/* number of live tuples */
		 (double) tuple_len / (1024 * 1024),		/* live tuples in MB */
		 tuple_percent,			/* live tuples in % */

		 (double) dead_tuple_count,	/* number of dead tuples */
		 (double) dead_tuple_len / (1024 * 1024), /* dead tuples in MB */
		 dead_tuple_percent,	/* dead tuples in % */

		 (double) free_space / (1024 * 1024), /* free/available space in
											   * MB */

		 free_percent,			/* free/available space in % */

		 /* overhead in % */
		 (nblocks == 0) ? 0.0 : 100.0
		 - tuple_percent
		 - dead_tuple_percent
		 - free_percent);

	PG_RETURN_FLOAT8(dead_tuple_percent);
}

#include "catalog/pg_index.h"
/*
 * flowig is copied from catalog/index.c
 *
 * IndexGetRelation: given an index's relation OID, get the OID of the
 * relation it is an index on.	Uses the system cache.
 */
static Oid
IndexGetRelation(Oid indexId)
{
	HeapTuple	tuple;
	Form_pg_index index;
	Oid			result;

	tuple = SearchSysCache(INDEXRELID,
						   ObjectIdGetDatum(indexId),
						   0, 0, 0);
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "IndexGetRelation: can't find index id %u",
			 indexId);
	index = (Form_pg_index) GETSTRUCT(tuple);
	Assert(index->indexrelid == indexId);

	result = index->indrelid;
	ReleaseSysCache(tuple);
	return result;
}
